import unittest

from TSPLifesaver.structure import *


class TestStructure(unittest.TestCase):
    def test_point(self):
        self.point_1 = PointWithEuclideanDistance([1, 1, 1, 1], name="test name 1")
        self.point_2 = PointWithEuclideanDistance([2, 2, 2, 2], name="test name 2")

        self.assertEqual(self.point_1.distance_to(self.point_2), 2)
        self.assertEqual(self.point_1.distance_to([2, 2, 2, 2]), 2)

        self.assertEqual(len(self.point_1), 4)
        self.assertEqual(self.point_1[0], 1)

    def test_route(self):
        self.points = [PointWithEuclideanDistance(i, i) for i in [[1, 1], [2, 2], [3, 3]]]
        self.route = BasicRoute(self.points, name="test route")
        self.assertEqual(self.route.distance(), sqrt(2)*4)



from typing import Iterable, MutableSequence, Type
from random import shuffle
from copy import deepcopy

from TSPLifesaver.abc import AbstractRoute, AbstractPoint
from TSPLifesaver.structure import BasicRoute, PointWithEuclideanDistance
from TSPLifesaver.optimizer import SimulatedAnnealing


def route_from_sequence(sequence: Iterable[MutableSequence], route: AbstractRoute = BasicRoute([]),
                        point_class: Type[AbstractPoint] = PointWithEuclideanDistance,
                        name_offset: int = 1, ) -> AbstractRoute:
    """
    :param route: Instances of the AbstractRoute class or its subclasses, defaults to empty instance of BasicRoute
    :param name_offset: Index of the name
    :param sequence: Sequence containing coordinates
    :param point_class: AbstractPoint or its subclasses ,defaults to PointWithEuclideanDistance
    :return: a new route
    """
    index = name_offset

    for pos in sequence:
        try:
            point = point_class(pos, name=f"{index}")
        except:
            point = point_class(pos)

        route.append(point)
        index += 1

    return route


def simulated_annealing(route: AbstractRoute, epoch: int = 100, temperature: float = 10000,
                        cooling_rate: float = 0.03, min_temperature: float = 1,
                        log: bool = False) -> AbstractRoute:
    """
    :param route: Instances of the AbstractRoute class or its subclasses
    :param epoch: Number of epochs to simulate, defaults to 100
    :param temperature: Temperature of the annealing, defaults to 10000
    :param cooling_rate: Cooling rate of the annealing, defaults to 0.03
    :param min_temperature: Minimum temperature of the annealing, defaults to 1
    :param log: Whether to print the log of the annealing, defaults to False
    :return: optimized route
    """
    if len(route):
        best_route = deepcopy(route)
        for i in range(epoch):
            if log:
                print(f"Running epoch {i} of {epoch}")
            shuffle(route)
            opt = SimulatedAnnealing(route, temperature=temperature,
                                     cooling_rate=cooling_rate, min_temperature=min_temperature)
            current_route = opt.optimize()
            if current_route.distance() < best_route.distance():
                best_route = deepcopy(current_route)
        return best_route
    else:
        return route


if __name__ == "__main__":
    print(simulated_annealing(route_from_sequence([[1, 1], [0, 0], [1, 0], [0, 1]]),log=True))



from typing import List, Any, MutableSequence, Sequence
from math import sqrt

from TSPLifesaver.abc import AbstractRoute, AbstractPoint


class BasicPoint(AbstractPoint):
    def __init__(self, pos: MutableSequence, name: Any = None):
        self.pos = pos
        self._name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    def __getitem__(self, item):
        return self.pos[item]

    def __setitem__(self, item, value):
        self.pos[item] = value

    def __delitem__(self, item):
        del self.pos[item]

    def __len__(self):
        return len(self.pos)

    def __str__(self):
        return str(self.name) + f"({str(self.pos)})"

    def distance_to(self, other: MutableSequence):
        """
        Calculates the Euclidean distance between two points.
        :param other:
        :return Number:
        """
        if len(other) != len(self.pos):
            raise ValueError("The dimensions of the two points do not match.")
        else:
            return sqrt(sum([(self.pos[i] - other[i]) ** 2 for i in range(len(self))]))


class PointWithEuclideanDistance(BasicPoint):
    def __init__(self, pos: MutableSequence, name: Any = None):
        super().__init__(pos, name)


class BasicRoute(AbstractRoute):
    def __init__(self, points: MutableSequence[AbstractPoint], name="BasicRoute"):
        self.points = points
        self.name = name

    def __iter__(self):
        return iter(self.points)

    def __getitem__(self, item):
        return self.points[item]

    def __setitem__(self, item, value):
        self.points[item] = value

    def __delitem__(self, item):
        del self.points[item]

    def __len__(self):
        return len(self.points)

    def __str__(self):
        string = self.name + "(\n"
        for point in self.points:
            string += f"{point.name}: {[point[i] for i in range(len(point))]}\n"
        string += ")"
        return string

    def insert(self, index, value):
        self.points.insert(index, value)

    def distance(self):
        """
        Calculates the total distance.
        :return:
        """
        return sum([pre.distance_to(after) for pre, after in zip(self[:-1], self[1:])])

    def swap(self, index_1: int, index_2: int) -> None:
        """
        Swaps two points
        :param index_1:
        :param index_2:
        :return:
        """
        self[index_1], self[index_2] = self[index_2], self[index_1]

    def append(self, value: AbstractPoint):
        self.points.append(value)


class CircularRoute(BasicRoute):
    def __init__(self, points: MutableSequence[AbstractPoint], name="CircularRoute"):
        super().__init__(points, name)

    def distance(self):
        return super().distance() + self[-1].distance_to(self[0])



from collections.abc import MutableSequence
from abc import ABC, abstractmethod
from numbers import Number


class AbstractPoint(ABC, MutableSequence):
    def __delitem__(self, key): ...

    def insert(self, index, value): ...

    @abstractmethod
    def __init__(self,pos):
        """
        Init the Point
        :param pos:
        """

    @property
    def name(self):
        """
        The name of the Point.
        :return: Any
        """
        return None

    @abstractmethod
    def distance_to(self, other: MutableSequence):
        """
        Calculate the distance between this Point and another.
        :param other:
        :return The distance between the:
        """


class AbstractRoute(ABC, MutableSequence):
    @abstractmethod
    def swap(self, index_1: int, index_2: int) -> None:
        """
        This method should swap the positions of the two elements by indexes.
        """

    @abstractmethod
    def distance(self):
        """
        This method should return the total length of the route.
        :return Number: The total length of the route:
        """


class AbstractOptimizer(ABC):
    @abstractmethod
    def optimize(self):
        """
        This method should start the process of optimization and return the result.
        """



import unittest

from TSPLifesaver.optimizer import *
from TSPLifesaver.structure import BasicRoute, PointWithEuclideanDistance


class TestStructure(unittest.TestCase):

    def test_SimulatedAnnealing(self):
        self.points = [PointWithEuclideanDistance(i) for i in
                       [[565, 575], [25, 185], [345, 750], [945, 685], [845, 655], [880, 660],
                        [25, 230], [525, 1000], [580, 1175], [650, 1130], [1605, 620], [1220, 580], [1465, 200],
                        [1530, 5],
                        [845, 680], [725, 370]]]
        self.route = BasicRoute(self.points, name="test route")
        optimizer = SimulatedAnnealing(self.route, 10000, 0.003, 1)
        self.route = optimizer.optimize()
        print("\n")
        print(self.route.distance())

    def test_ExhaustiveIndexing(self):
        self.points = [PointWithEuclideanDistance(i) for i in
                       [[565, 575], [25, 185], [345, 750], [945, 685], [845, 655], [880, 660],
                        [25, 230], [525, 1000]]]
        self.route = BasicRoute(self.points, name="test route")
        optimizer = ExhaustiveIndexing(self.route)
        self.route = optimizer.optimize()

        print("\n")
        print(self.route.distance())



import random
import math
from copy import deepcopy

from TSPLifesaver.abc import AbstractRoute, AbstractPoint
from TSPLifesaver.abc.abc import AbstractOptimizer


class SimulatedAnnealing(AbstractOptimizer):
    def __init__(self, initial_route: AbstractRoute, temperature, cooling_rate, min_temperature):
        """
        :param initial_route:
        :param initial_route:
        :param temperature:
        :param cooling_rate:
        :param min_temperature:
        """
        self.current_route = deepcopy(initial_route)
        self.best_route = deepcopy(initial_route)
        self.temperature = temperature
        self.cooling_rate = cooling_rate
        self.min_temperature = min_temperature

    def optimize(self):
        while self.temperature > self.min_temperature:
            new_route = deepcopy(self.current_route)

            # exchange randomly
            i, j = random.sample(range(len(new_route)), 2)
            new_route.swap(i, j)

            # calc cost
            current_cost = self.current_route.distance()
            new_cost = new_route.distance()
            cost_difference = current_cost - new_cost

            # accepting the new result?
            if cost_difference > 0 or math.exp(cost_difference / self.temperature) > random.random():
                self.current_route = new_route
                if new_cost < self.best_route.distance():
                    self.best_route = new_route

            # decrease the temperature
            self.temperature *= (1 - self.cooling_rate)

        return self.best_route


class ExhaustiveIndexing(AbstractOptimizer):
    def __init__(self, initial_route: AbstractRoute, max_iterations: int = None):
        self.best_route = deepcopy(initial_route)
        self.max_iterations = max_iterations
        self.count = 0

    def _permute(self, route: AbstractRoute, current_index=0):
        if current_index == len(route):
            yield route

        else:
            for i in range(current_index, len(route)):
                route.swap(current_index, i)
                yield from self._permute(route, current_index + 1)
                route.swap(current_index, i)

    def optimize(self):
        for route in self._permute(deepcopy(self.best_route)):
            # print("length of current route: ", route.length())
            if self.max_iterations is not None and self.count >= self.max_iterations:
                break

            if route.distance() < self.best_route.distance():
                self.best_route = deepcopy(route)
            self.count += 1

        return self.best_route
